#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <set>
#include <cassert>
#include <iostream>
#include <fstream>

using std::cout;
using std::cerr;
using std::endl;

const double PI = 3.141592653589793238462;
const double EPS = 1e-9;
const double INF = 1e100;
const double STEP = 1.0 / 60.0;

template<typename T>
T sqr(const T& t) {
	return t * t;
}
template<typename T>
T min(const T& a, const T& b) {
	if (a < b) return a;
	return b;
}


struct Point {
	double x, y;
	Point(): x(0), y(0)
	{ }
	Point(double x_, double y_): x(x_), y(y_) 
	{ }
	Point operator * (double num) const {
		return Point(x * num, y * num);
	}
	Point operator / (double num) const {
		return Point(x / num, y / num);
	}
};

struct MovableCircle {
	MovableCircle(const Point& pos_, const Point& velocity_, double radius_):
		pos(pos_), velocity(velocity_), radius(radius_)
	{ }
	Point pos;
	Point velocity;
	double radius;
};

Point operator - (const Point& lhs, const Point& rhs) {
	return Point(lhs.x - rhs.x, lhs.y - rhs.y);
}

Point operator + (const Point& lhs, const Point& rhs) {
	return Point(lhs.x + rhs.x, lhs.y + rhs.y);
}


double distance(const Point& lhs, const Point& rhs) {
	return std::sqrt(sqr(lhs.x - rhs.x) + sqr(lhs.y - rhs.y));
}

double innerProduct(const Point& lhs, const Point& rhs) {
	return lhs.x * rhs.x + lhs.y * rhs.y;
}

Point normilize(const Point& p) {
	double norm = distance(Point(), p);
	if (abs(norm) > EPS) {
		return p / norm;
	}
	return p;
}

double intersection(const MovableCircle& lhs, const MovableCircle& rhs) {
	Point speed = rhs.velocity - lhs.velocity;
	Point pos = rhs.pos - lhs.pos;
	double radius = lhs.radius + rhs.radius;

	if (distance(pos, Point()) < radius) {
		return 0.0;
	}


	double a = innerProduct(speed, speed);
    if (a < EPS) return INF;
	double b = 2 * innerProduct(speed, pos);
	double c = innerProduct(pos, pos) - sqr(radius);
	double d = b * b - 4.0 * a * c;
	if (d < -EPS) {
		return INF;
	}
	else {
		double t;
		t = (-b - std::sqrt(abs(d))) / (2.0 * a);
		if (t > EPS) {
			return t;
		}
		t = (-b + std::sqrt(abs(d))) / (2.0 * a);
		if (t > EPS) {
			return t;
		}
		return INF;
	}
}

double dangerous(const World* world, const Point& position) {
	// Than greater than worser
	double d = 0.0;
	for (size_t i = 0; i < world->MAXMONSTERS; ++i) {
		if (world->monsters[i].ClassName == -1) continue;
		Point monster(world->monsters[i].PositionX, world->monsters[i].PositionY);
        double dist = /*world->monsters[i].BiteDamage * */std::max(0.0, distance(position, monster) - world->player.Radius - world->monsters[i].Radius);
        d += 1.0 / (EPS + dist);
	}

	d += 1.0 / (EPS + min(position.x, 1.0 - position.x));
	d += 1.0 / (EPS + min(position.y, 1.0 - position.y));
	
	Point corners[4] = {
		Point(0.0, 0.0),
		Point(1.0, 0.0),
		Point(0.0, 1.0),
		Point(1.0, 1.0),
	};
	for (int i = 0; i < 4; ++i) {
		d += 1.0 / (EPS + distance(position, corners[i]));
	}
	return d;
}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;

	//cerr << "Gun cooldown period: " << world->player.GunCooldownPeriod << endl;

	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();


		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		Point position(world->player.PositionX, world->player.PositionY);

		fprintf(stderr, "frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		cerr << "Finding move..." << endl;

        double bestDangerous = INF;
        Point bestDirection;
        for (double alpha = 0.0; alpha < 2 * PI; alpha += 0.1) {
            const double LENGTH = 20 * STEP;
            const double S = 2 * STEP;
            double d = 0.0;
			Point direction(cos(alpha), sin(alpha));
    		for (int i = 0; i < world->MAXMISSILES; ++i) {
    			if (world->missiles[i].ClassName == -1) continue;
			    MovableCircle missle(Point(world->missiles[i].PositionX, world->missiles[i].PositionY), 
								    Point(world->missiles[i].VelocityX, world->missiles[i].VelocityY), world->missiles[i].Radius);
                MovableCircle player(position, direction * world->player.MaxVelocity, world->player.Radius);
                double t = intersection(missle, player);
                if (t < 0.1) {
                    continue;
                }
                d += 100.0 * (world->missiles[i].Damage / (EPS + t));
            }
            cerr << "d1 = " << d << endl;
            for (double step = 0.0; step <= LENGTH;) {
                double dx = ((direction.x > 0.0) ? (1.0 - position.x) : position.x) / std::abs(direction.x);
                double dy = ((direction.y > 0.0) ? (1.0 - position.y) : position.y) / std::abs(direction.y);
                double realStep = min(step, min(dx, dy));

                Point newPos = position + direction * realStep;
                //if (newPos.x < 0.0 || newPos.x > 1.0 || newPos.y < 0.0 || newPos.y > 1.0) {
                //    d += INF;
                //}
                //else {
                double dd = (LENGTH - step) * std::log(1.0 + dangerous(world, newPos));
                //cerr << dd << " ";
                d += dd;
                //std::sqrt(LENGTH - step) * std::log(1.0 + dangerous(world, newPos));                
                //}
                if (step < LENGTH / 4.0) {
                    step += S;
                }
                else {
                    step += 2 * S;
                }
            }
            cerr << endl;
            cerr << "d2 = " << d << endl;
            if (d < bestDangerous) {
                bestDangerous = d;
                bestDirection = direction;
            }
        }

		//const size_t D = 9;
		//double dx[D] = {-1, -1, -1, 0, 1, 1, 1, 0, 0}; 
		//double dy[D] = {-1, 0, 1, 1, 1, 0, -1, -1, 0};
  //      int pos = -1;
		//double dang = INF;
		//for (int i = 0; i < D; ++i) {
		//	Point newPos(position + normilize(Point(dx[i], dy[i])) * 10 * STEP);
		//	if (newPos.x < 0.0 || newPos.x > 1.0 || newPos.y < 0.0 || newPos.y > 1.0) {
		//		continue;
		//	}
		//	double d = dangerous(world, newPos);
		//	if (d < dang) {
		//		pos = i;
		//		dang = d;
		//	}
		//}
        Point newPosition = position + (normilize(bestDirection) * STEP);

		world->run(bestDirection.x, bestDirection.y);

		cerr << "Finding shoot..." << endl;

		double health[world->MAXMONSTERS] = {};
		Point monsters[world->MAXMONSTERS] = {};
		for (int i = 0; i < world->MAXMONSTERS; ++i) {
			if (world->monsters[i].ClassName == -1) continue;
			health[i] = world->monsters[i].HitPoints;
			monsters[i] = Point(world->monsters[i].PositionX, world->monsters[i].PositionY);
		}

		for (int i = 0; i < world->MAXMISSILES; ++i) {
			if (world->missiles[i].ClassName == -1) continue;

			//cerr << "Missle = (" << world->missiles[i].PositionX << ", " << world->missiles[i].PositionY << ")" 
			//	<< "(" << world->missiles[i].VelocityX << ", " << world->missiles[i].VelocityY << ")" << endl;
		
			int firstMonster = -1;
			double minTime = INF;
			MovableCircle missle(Point(world->missiles[i].PositionX, world->missiles[i].PositionY), 
								Point(world->missiles[i].VelocityX, world->missiles[i].VelocityY), world->missiles[i].Radius);

			for (int j = 0; j < world->MAXMONSTERS; ++j) {
				if (world->monsters[j].ClassName == -1) continue;

				MovableCircle monster(Point(monsters[j]), 
								Point(world->monsters[j].VelocityX, world->monsters[j].VelocityY), world->monsters[j].Radius);
				double time = intersection(missle, monster);
				if (minTime > time + EPS) {
					time = minTime;
					firstMonster = j;
				}
			}
			//cerr << "firstMonster = " << firstMonster << endl;
			health[firstMonster] -= world->missiles[i].Damage;
		}

        //std::set<int> monstersTypes;

		int nearestMonster = -1;
		double minDist = INF;
        bool isBullet = false;
		for (int i = 0; i < world->MAXMONSTERS; ++i) {
			if (world->monsters[i].ClassName == -1 || health[i] < 0.0) continue;
			double dist = distance(monsters[i], position);

            if (dist > 2 * STEP && isBullet && std::abs(world->monsters[i].BulletDamage) < EPS) {
                continue;
            }

            bool newBullet = std::abs(world->monsters[i].BulletDamage) > EPS;
			if (isBullet != newBullet || dist < minDist) {
				minDist = dist;
				nearestMonster = i;
                isBullet = newBullet;
			}
		}
        if (nearestMonster == -1) {
            continue;
        }


        //cerr << "Types = {";
        //for (std::set<int>::iterator it = monstersTypes.begin(); it != monstersTypes.end(); ++it) {
        //    cerr << *it << " ";
        //}
        //cerr << "}" << endl;

		double SHOOT_STEP = 0.01;
        bool ok = false;
        cerr << newPosition.x << " " << newPosition.y << endl;
		MovableCircle monster(Point(monsters[nearestMonster]), 
			Point(world->monsters[nearestMonster].VelocityX, world->monsters[nearestMonster].VelocityY), world->monsters[nearestMonster].Radius);
		for (double alpha = 0.0; alpha <= 2 * PI && !ok; alpha += 0.01) {
			Point direction(cos(alpha), sin(alpha));
			MovableCircle missile(position, direction * world->player.BulletVelocity, world->player.BulletRadius);
			if (intersection(missile, monster) < INF) {
				world->shoot(direction.x, direction.y);
				ok = true;
			}
		}
        if (!ok) { 
            Point shoot = monsters[nearestMonster] - newPosition;
			world->shoot(shoot.x, shoot.y);
        }


		////Choose the first monster that is still alive
		//for ( int j = 0; j < world->MAXMONSTERS; j++)
		//	if (world->monsters[j].ClassName >= 0) {
		//		//Shoot the bastard! (Shoot in the direction of the monster)
		//		world->shoot(world->monsters[j].PositionX - world->player.PositionX,
	    //				   world->monsters[j].PositionY - world->player.PositionY);
		//		break;
		//	}
	}

	return 0;
}
