import static java.lang.Math.PI;
import static java.lang.Math.atan2;
import static java.lang.Math.cos;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.sin;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Random;

class Task {
	static class Vector {
		double x, y;
		
		Vector(double x, double y) {
			this.x = x;
			this.y = y;
		}
		
		Vector sub(Vector o) {
			return new Vector(x - o.x, y - o.y);
		}
		
		Vector add(Vector o) {
			return new Vector(x + o.x, y + o.y);
		}
		
		Vector mul(double o) {
			return new Vector(x * o, y * o);
		}
		
		double len2() {
			return x * x + y * y;
		}
		
		double len() {
			return Math.sqrt(len2());
		}
		
		Vector norm() {
			return mul(1. / len());
		}

		public double angle() {
			if (len() < 1e-8) {
				return 0.;
			}
			return atan2(y, x);
		}

		public Vector clip() {
			return new Vector(max(0., min(1., x)), max(0., min(1., y)));
		}
	}
	
	static class Monster {
		CMonster cmonster;
		Vector pos, vel;
		double radius;
		double hits;
		boolean shooting;
		
		Monster(CMonster cmonster) {
			this.cmonster = cmonster;
			this.pos = new Vector(cmonster.PositionX, cmonster.PositionY);
			this.vel = new Vector(cmonster.VelocityX, cmonster.VelocityY);
			this.radius = cmonster.Radius;
			this.hits = cmonster.HitPoints;
			this.shooting = cmonster.BulletDamage > 0;
		}
	}

	private static World world;
	private static Monster[] aliveMonsters;
	private static Vector playerPos;
	
	static double sMul(Vector a, Vector b) {
		return a.x * b.x + a.y * b.y;
	}
	
	static double vMul(Vector a, Vector b) {
		return a.x * b.y - a.y * b.x;
	}
	
	static int cmp(double a, double b) {
		final double EPS = 1e-8;
		if (Math.abs(a - b) < EPS) {
			return 0;
		}
		return a < b ? -1 : 1;
	}
	
	static Vector shootDir(Vector p, Vector v, double vb) {
		double A = v.len2() - vb * vb;
		double B = 2. * sMul(p, v);
		double C = p.len2();
		double D = B * B - 4 * A * C;
		if (cmp(D, 0) < 0) {
			return null;
		}
		D = Math.sqrt(Math.max(0., D));
		double t1 = (-B - D) / 2. / A;
		double t2 = (-B + D) / 2. / A;
		if (cmp(Math.max(t1, t2), 0) < 0) {
			return null;
		}
		double t = minNonNegative(t1, t2);
		if (Double.isNaN(t)) {
			return null;
		}
		return p.add(v.mul(t));
	}
	
	static double minNonNegative(double t1, double t2) {
		if (cmp(Math.max(t1, t2), 0) < 0) {
			return Double.NaN;
		}
		return cmp(t2, 0) < 0 ? t1 : cmp(t1, 0) < 0 ? t2 : Math.min(t1, t2);
	}
	
	final static Vector[] angle = {new Vector(0., 0.), new Vector(0., 1.), new Vector(1., 0.), new Vector(1., 1.)};
	
	public static void main(String[] args) throws IOException{

		//Check for enough parameters
		if (args.length < 2){
			System.err.println("parameters: IP port");
			System.exit(10);
		}

		world = new World(args[0], Integer.parseInt(args[1]));

		makeMove();
		Random rnd = new Random(42);
		System.err.println("damage = " + world.player.BulletDamage);
		System.err.println("bullet hit points = " + world.player.BulletHitPoints);
		System.err.println("bullet velocity = " + world.player.BulletVelocity);
		System.err.println("bullet radius = " + world.player.BulletRadius);
		System.err.println("velocity = " + world.player.MaxVelocity);

		int lastFrame = 0;
		while (true) {
			if (world.player.Time > lastFrame + 1) {
				System.err.println("Warning: skipped frame #" + lastFrame + 1);
			}
			lastFrame = world.player.Time;
			System.out.println("frame : " + world.player.Time + " HP: "+ world.player.HitPoints + " x: " + world.player.PositionX + " y: " + world.player.PositionY); 


			final int DIRECTIONS = 24;
			final double[] STEPS = {0.5, 1., 1.5, 1.7, 2.0};
			Vector center;
			if (3200 < world.player.Time && world.player.Time <= 3500) {
				center = new Vector(0.5, 0.75);
			} else if (3500 < world.player.Time && world.player.Time <= 3800) {
				center = new Vector(0.5, 0.25);
			} else {
				center = new Vector(0.5, 0.5);
			}
			final double dirToCenter = center.sub(playerPos).angle();
			Vector bestDirection = null;
			double bestDirectionDamage = Double.POSITIVE_INFINITY;
			for (int dir = 0; dir < DIRECTIONS; ++dir) {
				double alpha = 2. * PI / DIRECTIONS * dir + dirToCenter;
				double maxDamage = 0.;
				Vector dirv = new Vector(cos(alpha), sin(alpha));
				for (double step : STEPS) {
					Vector newPos = playerPos.add(dirv.mul(step * world.player.MaxVelocity)).clip();
					double damage = 0.;
					for (Monster m : aliveMonsters) {
						if (dist(m.pos, newPos) < m.cmonster.MaxVelocity * step) {
							damage += m.cmonster.BiteDamage;
						}
					}
					maxDamage += damage;
				}
				maxDamage /= STEPS.length;
				for (CMissile cmissile : world.missiles) {
					if (cmissile.ClassName < 0) {
						continue;
					}
					double t = hitTime(dirv.mul(world.player.MaxVelocity), playerPos, world.player.Radius, cmissile);
					if (!Double.isNaN(t) && t < 0.3) {
						maxDamage += cmissile.Damage / t;
					}
				}
				// 
				Vector newPos = playerPos.add(dirv.mul(0.2 * world.player.MaxVelocity)).clip();
				double distToCenter = dist(new Vector(0.5, 0.5), newPos);
				if (distToCenter > 0.60) {
					maxDamage *= 1.0 + 30 * (distToCenter - 0.60);
				}
				if (maxDamage < bestDirectionDamage) {
					bestDirectionDamage = maxDamage;
					bestDirection = dirv;
				}
			}
			Vector curDir = bestDirection;
			moveTo(curDir);
			
			final Vector newPos = playerPos.add(curDir.mul(0.1 * world.player.MaxVelocity));
			
			
//			Vector mCenter = new Vector(0., 0.);
//			for (Monster m : aliveMonsters) {
//				mCenter = mCenter.add(m.pos);
//			}
//			mCenter.x /= aliveMonsters.length;
//			mCenter.y /= aliveMonsters.length;
//			moveTo(playerPos.sub(mCenter));
			
			for (CMissile cmissile : world.missiles) {
				if (cmissile.ClassName < 0) {
					continue;
				}
				Monster firstHit = null;
				double firstHitT = Double.POSITIVE_INFINITY;
				for (Monster monster : aliveMonsters) {
					double t = hitTime(monster.vel, monster.pos, monster.radius, cmissile);
					if (!Double.isNaN(t) && t < firstHitT) {
						firstHit = monster;
						firstHitT = t;
					}
				}
				if (firstHit != null) {
					firstHit.hits -= cmissile.Damage;
				}
			}
			
			Comparator<Monster> bestMonster = new Comparator<Task.Monster>() {
				public int compare(Monster a, Monster b) {
					final double DANGER_RADIUS = 0.3;
					if (a == null) {
						return 1;
					}
					if (b == null) {
						return -1;
					}
					boolean dangera = distToS(a) < DANGER_RADIUS;
					boolean dangerb = distToS(b) < DANGER_RADIUS;
					if (dangera || dangerb || cmp(a.hits, b.hits) == 0) {
						return cmp(distToS(a), distToS(b));
					} else {
						if (a.shooting != b.shooting) {
							return a.shooting ? -1 : 1;
						}
						return cmp(a.hits, b.hits);
					}
				}

				private double distToS(Monster a) {
					return (a.shooting ? 0.5 : 1.) * dist(a.pos, newPos);
				}
			};
			
			Monster shootAt = null;
			for (Monster monster : aliveMonsters) {
				if (monster.hits > 0 && bestMonster.compare(monster, shootAt) < 0) {
					shootAt = monster;
				}
			}
			if (shootAt != null) {
				CPlayer player = world.player;
				Vector shootDir = shootDir(shootAt.pos.sub(new Vector(player.PositionX, player.PositionY)), shootAt.vel, world.player.BulletVelocity);
				shoot(shootDir);
			} else {
				shoot(new Vector(rnd.nextDouble() - 0.5, rnd.nextDouble() - 0.5));
			}
			makeMove();
		}
	}
	
	static double hitTime(Vector vel, Vector pos, double radius, CMissile cmissile) {
		Vector missilePos = new Vector(cmissile.PositionX, cmissile.PositionY);
		Vector missileV = new Vector(cmissile.VelocityX, cmissile.VelocityY);
		double missileRadius = cmissile.Radius;
		Vector v = missileV.sub(vel);
		Vector p = missilePos.sub(pos);
		double r = radius + missileRadius;
		double A = v.len2();
		double B = 2. * sMul(p, v);
		double C = p.len2() - r * r;
		double D = B * B - 4 * A * C;
		if (cmp(D, 0) < 0) {
			return Double.NaN;
		}
		D = Math.sqrt(Math.max(0., D));
		return minNonNegative((-B - D) / 2. / A, (-B + D) / 2. / A);
	}
	
	private static double dist(Vector a, Vector b) {
		return a.sub(b).len();
	}

	static void makeMove() {
		world.makeMove();
		ArrayList<Monster> list = new ArrayList<Monster>();
		for ( int i = 0; i < World.MAXMONSTERS; i++) {
			if (world.monsters[i].ClassName > 0.){
				list.add(new Monster(world.monsters[i]));
			}
		}
		aliveMonsters = (Monster[]) list.toArray(new Monster[list.size()]);
		playerPos = new Vector(world.player.PositionX, world.player.PositionY);
	}

	private static void moveTo(Vector v) {
		world.run((float)v.x, (float)v.y);
	}

	private static double distTo(Monster monster) {
		return playerPos.sub(monster.pos).len();
	}

	private static void shoot(Vector v) {
		world.shoot((float)v.x, (float)v.y);
	}
}
