#include "GameWorld.h"

GameWorld::GameWorld() : director(0), frags(0) {
	for (int i = 0; i<entities.size(); i++) entities[i] = 0;

	Entity **ptr = &entities[0];
	players = (Player**)ptr;
	ptr += MAX_PLAYER_COUNT;
	monsters = (Monster**)ptr;
	ptr += MAX_MONSTER_COUNT;
	missiles = (Missile**)ptr;
	ptr += MAX_MISSILE_COUNT;

	director = std::auto_ptr<Director>(new Director(this));
	director->Init();
}

GameWorld::~GameWorld() {
	for (int i = 0; i<entities.size(); i++) if (entities[i]) {
		delete entities[i];
		entities[i] = 0;
	}
}

//-------------------typeinfo, create/destroy...-----------------

bool IsPlayer(int &i) {
	return i>=0 && i<MAX_PLAYER_COUNT;
}
bool IsMonster(int &i) {
	i -= MAX_PLAYER_COUNT;
	return i>=0 && i<MAX_MONSTER_COUNT;
}
bool IsMissile(int &i) {
	i -= MAX_PLAYER_COUNT + MAX_MONSTER_COUNT;
	return i>=0 && i<MAX_MISSILE_COUNT;
}
bool IsPlayer(const Entity *ent) { int ID = ent->ID;  return IsPlayer(ID); }
bool IsMonster(const Entity *ent) { int ID = ent->ID;  return IsMonster(ID); }
bool IsMissile(const Entity *ent) { int ID = ent->ID;  return IsMissile(ID); }

int GameWorld::AllocateMonster() {
	for (int i = 0; i<MAX_MONSTER_COUNT; i++) if (!monsters[i])
		return int((Entity**)&monsters[i] - &entities[0]);
	return -1;
}
int GameWorld::AllocateMissile() {
	for (int i = 0; i<MAX_MISSILE_COUNT; i++) if (!missiles[i])
		return int((Entity**)&missiles[i] - &entities[0]);
	return -1;
}

Entity *GameWorld::BindEntity(int ID, Entity *ent) {
	if (ID < 0) {
		delete ent;
		return 0;
	}
	else {
		entities[ID] = ent;
		ent->ID = ID;
		ent->owner = this;
		return ent;
	}
}
void GameWorld::DestroyEntity(Entity *ent) {
	int ID = ent->ID;
	delete ent;
	entities[ID] = 0;
}

//--------------------collision optimization-----------------

void GameWorld::FillGrid() {
	grid.Clear();
	for (int i = 0; i<entities.size(); i++) if (entities[i]) {
		MovingCircle go = entities[i]->GetGeometry();
		std::vector<int2> cells = GetGridCells(go);
		for (int j = 0; j<cells.size(); j++)
			grid[cells[j].x][cells[j].y].push_back(entities[i]);
	}
}

GameWorld::EntitySet GameWorld::FindCollision(const MovingCircle &go) {
	std::vector<int2> cells = GetGridCells(go);

	//direct hash with level
	static int UsedArray[MAX_ENTITY_COUNT] = {0};
	static int UsedLevel = 0;
	static const int MAX_USED_LEVEL = 1000000000;
	if (UsedLevel == MAX_USED_LEVEL) {
		memset(UsedArray, 0, sizeof(UsedArray));
		UsedLevel = 0;
	}
	UsedLevel++;
	EntitySet collset;
	for (int j = 0; j<cells.size(); j++) {
		const Grid::Cell &c = grid[cells[j].x][cells[j].y];
		for (int u = 0; u<c.size(); u++) {
			int ID = c[u]->ID;
			if (UsedArray[ID] == UsedLevel) continue;
			UsedArray[ID] = UsedLevel;
			if (HasIntersection(go, c[u]->GetGeometry()))
				collset.push_back(c[u]);
		}
	}

/*	//vector + sort + unique
	static EntitySet res;
	res.clear();
	for (int j = 0; j<cells.size(); j++) {
		const Grid::Cell &c = grid[cells[j].x][cells[j].y];
		for (int u = 0; u<c.size(); u++) res.push_back(c[u]);
	}
	std::sort(res.begin(), res.end());
	res.resize(std::unique(res.begin(), res.end()) - res.begin());
	std::vector<Entity *> collset;
	collset.reserve(res.size());
	for (std::vector<Entity *>::iterator it = res.begin(); it != res.end(); it++)
		if (HasIntersection(go, (*it)->GetGeometry()))
			collset.push_back(*it);*/

/*	//set
	std::set<Entity *> res;
	for (int j = 0; j<cells.size(); j++) {
		const Grid::Cell &c = grid[cells[j].x][cells[j].y];
		for (int u = 0; u<c.size(); u++) res.insert(c[u]);
	}
	std::vector<Entity *> collset;
	collset.reserve(res.size());
	for (std::set<Entity *>::iterator it = res.begin(); it != res.end(); it++)
		if (HasIntersection(go, (*it)->GetGeometry()))
			collset.push_back(*it);*/

/*	//dull collision check
	std::vector<Entity *> checkset;
	for (int i = 0; i<entities.size(); i++) if (entities[i])
		if (HasIntersection(go, entities[i]->GetGeometry()))
			checkset.push_back(entities[i]);
	std::sort(collset.begin(), collset.end());
	std::sort(checkset.begin(), checkset.end());
	if (collset != checkset) {
		//OMG;
		delete this;
	}*/

	return collset;
}

//---------------------Iterate between frames-------------------------

int GameWorld::Iterate(double deltatime) {

//--------------pass time on all the timers (cooldowns)--------------
	for (int i = 0; i<entities.size(); i++) if (entities[i]) {
		Entity *ent = entities[i];
		ent->newposition = ent->position;
		ent->PassTime(deltatime);
	}
	//plus on director
	director->PassTime(deltatime);

//------------------------make AI/player work---------------------------------
	//fill grid with still objects
	FillGrid();
	//run each monster AI with set of entities nearby
#ifdef USE_OMP
	omp_set_num_threads(omp_get_num_procs());
	#pragma omp parallel for schedule(dynamic)
#endif
	for (int i = 0; i<MAX_MONSTER_COUNT; i++) if (monsters[i]) {
		Monster *mon = monsters[i];
		const double2 &ctr = mon->position;
		double rad = mon->ai->GetAwarenessRadius();
		MovingCircle vicinity(ctr, ctr, rad);

		EntitySet aware = FindCollision(vicinity);
		for (int j = 0; j<MAX_PLAYER_COUNT; j++) if (players[j])
			if (std::find(aware.begin(), aware.end(), players[j]) == aware.end())
				aware.push_back(players[j]);

		mon->ai->Think(aware);
	}

	//listen to player's needs
	if (playerwantstoshoot)
		players[0]->Shoot(playershotdirection);
	
//----------------------------move and collide-------------------------------
	//fill grid with moving objects
	for (int i = 0; i<entities.size(); i++) if (entities[i]) {
		Entity *ent = entities[i];
		ent->Move(deltatime);
	}
	FillGrid();

	//collide missiles with monsters and players
	for (int i = 0; i<MAX_MISSILE_COUNT; i++) if (missiles[i]) {
		Missile *miss = missiles[i];
		EntitySet collides = FindCollision(miss->GetGeometry());
		std::vector<Entity *> &nh = miss->nowhit;
		std::vector<Entity *> newnh;
		for (EntitySet::iterator it = collides.begin(); it != collides.end(); it++) {
			if (IsMonster(*it) || IsPlayer(*it)) {
				if (std::find(nh.begin(), nh.end(), *it) == nh.end())
					miss->DealDamage(*it);
				newnh.push_back(*it);
			}
		}
		nh = newnh;
		//check field borders
		if (!InField(miss->newposition)) miss->hitpoints = -1.0;
	}

	//collide player with monsters
	for (int i = 0; i<MAX_PLAYER_COUNT; i++) if (players[i]) {
		Player *pl = players[i];
		EntitySet collides = FindCollision(pl->GetGeometry());
		for (EntitySet::iterator it = collides.begin(); it != collides.end(); it++)
			if (IsMonster(*it)) {
				Monster *mon = (Monster*)(*it);
				MeleeWeapon *teeth = mon->teeth.get();
				if (teeth) teeth->Bite(pl);
			}
		//check field borders for player
		ForceToField(pl->newposition, pl->radius);
	}

//----------------------------apply position changes--------------------------------
	//apply moves and find deaths
	for (int i = 0; i<entities.size(); i++) if (entities[i]) {
		Entity *ent = entities[i];
		ent->position = ent->newposition;
		if (ent->hitpoints <= 0.0) {
			if (IsPlayer(ent))
				return DEATH;

			//If player did last hit
			if (IsMonster(ent) && ent->hitpoints <= -INF/2) frags++;

			DestroyEntity(ent);
		}
	}

	return SUCCESS;
}

int GameWorld::GetScore() {
	return frags;
}

void GameWorld::SetPlayerMove(double2 direction) {
	players[0]->SetMoveDirection(direction);
}

void GameWorld::SetPlayerShoot(double2 direction) {
	playerwantstoshoot = (!direction > EPS);
	playershotdirection = direction;
}
