#include "Serialization.h"
#include "../gamelogic/GameWorld.h"
#include <omp.h>
#include <algorithm>
#include <limits>

//wait for mutex (win-dependent)?
//or eat CPU cycles instead...
//#define LOW_CPU_USAGE


//do we always print stuff to stdout?
static const bool ALWAYS_PRINT_STATS = true;

//maximal gametime allowed - instadeath occurs when exceeded
static const double MAX_GAME_TIME = 300.0;
//equal to 300.0 on final testing
//equal to 70.0 on current testing

//this file contains contestant score
static const char SCORE_FILENAME[] = "score.txt";

//this file contains info for contestant
static const char INFO_FILENAME[] = "info.txt";

//this file contains server log
static const char LOG_FILENAME[] = "game.blog";

//default port to listen
static const int DEFAULT_PORT = 6666;

//player's HP is set to THIS when gameover happens 
static const double DEAD_PLAYER_HITPOINTS = -1.0;

//world with dead player will remain for THIS time
static const double DEAD_WORLD_TIME = 1.0;

//upper bound of FPS = lower bound on deltatime
static const double MAX_GAME_FPS = 60;

double GetTime() { return omp_get_wtime(); }

#ifdef LOW_CPU_USAGE
	#include <windows.h>
	HANDLE waitmutex;
	HANDLE waitevent;
#endif

World *networld, *netbuffer;
GameWorld *world;
double creationtime;

FILE *flog, *fscore, *finfo;

int framecount;
volatile int netsendcount;
double currenttime;
double deltatime, maxdeltatime, avgdeltatime;
double iterationtime, maxiterationtime, avgiterationtime;

void Terminate(const char *str, int exitcode) {
	fprintf(stderr, "%s\n", str);
	fflush(stderr);
	exit(exitcode);
}

void WriteFinal() {
	printf("Frame / NetSend / Loss = %d / %d / %d\n", framecount, netsendcount, framecount - netsendcount);
	printf("\tScore = %d\n", world->GetScore());
	printf("\tMax delta / logic time = %8.6lf / %8.6lf\n", maxdeltatime, maxiterationtime);
	printf("\tAvg delta / logic time = %8.6lf / %8.6lf\n", avgdeltatime, avgiterationtime);
}

void WriteIngame() {
	printf("\t    delta / logic time = %8.6lf / %8.6lf\n", deltatime, iterationtime);
	printf("\tPlayer HP X Y = %5.3lf %5.3lf %5.3lf\n", world->players[0]->hitpoints, world->players[0]->position.x, world->players[0]->position.y);

	//count living entities
	int lpls = 0;
	int lmons = 0;
	int lmiss = 0;
	for (int i = 0; i<world->entities.size(); i++) {
		Entity *ent = world->entities[i];
		if (!ent) continue;
		if (IsPlayer(ent)) lpls++;
		if (IsMonster(ent)) lmons++;
		if (IsMissile(ent)) lmiss++;
	}
	printf("Entities = %d (%d:%d:%d)\n", lpls+lmons+lmiss, lpls, lmons, lmiss);
}

void WriteStats(bool final) {
	if (final) {
		printf("\n");
		WriteFinal();
	}
	else {
		if (ALWAYS_PRINT_STATS) {
			WriteFinal();
			WriteIngame();
		}
		else {
			printf(".");
			fflush(stdout);
		}
	}
}

void WriteLog() {
	if (flog) {
		netbuffer->writeFrameToFile(flog);
		fflush(flog);
	}
}

void WriteScore(int score) {
	static const int MAXLEN = 15;
	char buff[MAXLEN + 1];
	sprintf(buff, "%d", score);
	for (int i = strlen(buff); i<MAXLEN; i++) buff[i] = ' ';
	buff[MAXLEN] = 0;

	if (ferror(fscore)) Terminate("Score file error", 666);
	rewind(fscore);
	fprintf(fscore, "%s", buff);
	fflush(fscore);
}
void WriteInfo() {
	if (ferror(finfo))  Terminate("Info file error", 666);
	rewind(finfo);
	fprintf(finfo, "Score = %d\n", world->GetScore());
	fprintf(finfo, "Frame / NetSend / Loss = %d / %d / %d\n", framecount, netsendcount, framecount - netsendcount);
	fprintf(finfo, "Avg delta time = %8.6lf\n", avgdeltatime);
	for (int i = 0; i<16; i++) fprintf(finfo, " ");
	fflush(finfo);
}

void GameIterationThread() {
	static const double FrameDeltaTime = 1.0 / MAX_GAME_FPS;
	framecount = 0;
	while (1) {
		//calc deltatime
		double newtime = GetTime();
		deltatime = newtime - currenttime;
		
		if (deltatime < FrameDeltaTime) {
#ifdef LOW_CPU_USAGE
			int sleeptime = int((FrameDeltaTime - deltatime) * 1000) - 1;
			if (sleeptime < 1) continue;
			WaitForSingleObject(waitmutex, sleeptime);
#endif
			continue;
		}
		currenttime = newtime;
		framecount++;

		#pragma omp critical (World)
		{
			world->SetPlayerMove(double2(netbuffer->player.RunX, netbuffer->player.RunY));
			world->SetPlayerShoot(double2(netbuffer->player.ShootX, netbuffer->player.ShootY));
		}

		//iterate
		double iterationstart = GetTime();
		bool gameover = (world->Iterate(FrameDeltaTime) != GameWorld::SUCCESS);
		if (world->director->gametime > MAX_GAME_TIME) gameover = true;
		iterationtime = GetTime() - iterationstart;

		//set negative player HP if death occurs
		if (gameover) world->players[0]->hitpoints = DEAD_PLAYER_HITPOINTS;
		//print score after iteration
		WriteScore(world->GetScore());

		//serialize world
		#pragma omp critical (World)
		{
			SerializeWorld(*netbuffer, world);
			netbuffer->player.Time = framecount;
#ifdef LOW_CPU_USAGE
			SetEvent(waitevent);
#endif
		}

		avgdeltatime = ((framecount-1) * avgdeltatime + deltatime) / framecount;
		avgiterationtime = ((framecount-1) * avgiterationtime + iterationtime) / framecount;
		if (framecount > 300) {
			if (maxdeltatime < deltatime) maxdeltatime = deltatime;
			if (maxiterationtime < iterationtime) maxiterationtime = iterationtime;
		}

		//write log and stats/info
		WriteInfo();
		WriteStats(false);
		WriteLog();

		if (gameover) {
			static double gameovertime = GetTime();
			printf("Waiting after death\n");
			fflush(stdout);
			while (GetTime() - gameovertime < DEAD_WORLD_TIME);
			WriteStats(true);
			Terminate("Game over", 0);          
		}
	}
}

bool IsBadFloat(const float &x) {
	try {
		if (x != x) return true;
		if (double(x) > 1E100 || double(x) < -1E100) return true;
	}
	catch (...) {
		return true;
	}
	return false;
}
void CheckVector(float &x, float &y) {
	if (IsBadFloat(x) || IsBadFloat(y)) {
		printf("Bad float output!\n");
		fflush(stdout);
		x = y = 0;
	}
}
void NetCommunicationThread() {
	netsendcount = 0;
	bool refresh = true;
	while (1) {
		//get player move
		if (refresh) {
			#pragma omp atomic
				netsendcount++;
			networld->getMove();
		}
#ifdef LOW_CPU_USAGE
		WaitForSingleObject(waitevent, INFINITE);
		ResetEvent(waitevent);
#endif
		#pragma omp critical (World)
		{
			CheckVector(networld->player.RunX, networld->player.RunY);
			CheckVector(networld->player.ShootX, networld->player.ShootY);
			netbuffer->player.RunX = networld->player.RunX;
			netbuffer->player.RunY = networld->player.RunY;
			netbuffer->player.ShootX = networld->player.ShootX;
			netbuffer->player.ShootY = networld->player.ShootY;
			refresh = (netbuffer->player.Time != networld->player.Time);
			if (refresh) memcpy(networld, netbuffer, sizeof(World));
		}
		if (refresh) {
			networld->sendWorld();
			if (networld->player.HitPoints < 0) refresh = false;
		}
	}
}

int main(int argc, char **argv) {
	fscore = fopen(SCORE_FILENAME, "wt");
	if (!fscore) Terminate("Cannot open score file", 666);
	finfo = fopen(INFO_FILENAME, "wt");
	if (!finfo) Terminate("Cannot open info file", 666);
	flog = fopen(LOG_FILENAME, "wb");
	WriteScore(-1);

	//set port
	int port = DEFAULT_PORT;
	int temp;
	if (argc >= 2 && sscanf(argv[1], "%d", &temp) == 1) port = temp;

	//set seed
	int seed = int(GetTime() * 1000.0);
	if (argc >= 3 && sscanf(argv[2], "%d", &temp) == 1) seed = temp;
	seed &= (1<<15) - 1;
	srand(seed);
	FILE *fseed = fopen("seed.txt", "wt");
	fprintf(fseed, "%d\n", seed);
	fclose(fseed);

    //create world (GOD's problem =) )
	world = new GameWorld();
	fprintf(stdout, "waiting for client!\n");  fflush(stdout);
	fprintf(stderr, "waiting for client!\n");  fflush(stderr);
	networld = new World(port);
	fprintf(stderr, "Connection established\n");
	netbuffer = (World*)malloc(sizeof(World));
	memcpy(netbuffer, networld, sizeof(World));
	currenttime = creationtime = GetTime();
	fprintf(stderr, "Game with seed = %d started\n", seed);

	#pragma omp parallel num_threads(2)
	{
		int t = omp_get_thread_num();
#ifdef LOW_CPU_USAGE
		if (t == 0) waitevent = CreateEvent(NULL, true, false, "Server_WaitEvent_068460");
		if (t == 1) waitmutex = CreateMutex(NULL, true, "Server_TimedWaitMutex_972786");
#endif
		#pragma omp barrier
		if (t == 0) {
			GameIterationThread();
		}
		else {
			NetCommunicationThread();
		}	
	}
}
