#include "../cpp/world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <vector>
#undef min
#undef max
#include <algorithm>

static const double EPS = 1e-9;
static const double INF = 1e+50;
static const double DELTATIME = 1.0 / 60.0;

struct Point {
	double x, y;
	
	inline Point() : x(0), y(0) {}
	inline Point(double _x, double _y) : x(_x), y(_y) {}

	inline       double &operator[] (int i)       { return (i==0 ? x : y); }
	inline const double &operator[] (int i) const { return (i==0 ? x : y); }

	inline Point operator+ (const Point &b) const { return Point(x+b.x, y+b.y); }
	inline Point operator- (const Point &b) const { return Point(x-b.x, y-b.y); }
	inline Point operator* (double b) const { return Point(x*b, y*b); }
	inline Point operator/ (double b) const { return *this * (1.0/b); }

	inline void operator+= (const Point &b) { *this = *this + b; }
	inline void operator-= (const Point &b) { *this = *this - b; }
	inline void operator*= (double b) { *this = *this * b; }
	inline void operator/= (double b) { *this = *this / b; }

	inline double operator^ (const Point &b) const { return x*b.x + y*b.y; }
	inline double operator% (const Point &b) const { return x*b.y - y*b.x; }

	inline double operator!() const { return sqrt(*this ^ *this); }
};

Point Normal(const Point &x) {
	Point res = x;
	double norm = !res;
	if (norm > EPS) res /= norm;
	return res;
}

inline double trandom() { return rand() / 32768.0; }

#define POINT(name) Point(name##X, name##Y)
#define COORDS(pnt) (pnt).x, (pnt.y)
static const int MONSTERS = World::MAXMONSTERS;
static const int MISSILES = World::MAXMISSILES;

extern World *world;


static const double MOVERADIUS = 0.4;
static const double SPREADTIME = 0.1;
static const double AWARERADIUS = 0.4;
static const double BITEWGT = 0.5;
static const double SHOOTANGLECOS = 0.70710678118654752440084436210485;

Point TargetMonsterDull(const CMonster &monster) {
	return POINT(monster.Position) - POINT(world->player.Position);
}

Point TargetMonsterMoving(const CMonster &monster) {
	Point pos = POINT(monster.Position) - POINT(world->player.Position);
	Point vel = POINT(monster.Velocity);
	pos += vel * (2.0*trandom()-1.0) * SPREADTIME;
	double speed = world->player.BulletVelocity;

	double qa = (vel ^ vel) - (speed * speed);
	double qb = (pos ^ vel);
	double qc = (pos ^ pos);
	if (qa > -EPS) return pos;
	double qd = qb*qb - qa*qc;
	qd = sqrt(qd);
	double t1 = (-qb - qd) / qa;
	double t2 = (-qb + qd) / qa;
	if (t1 < 0.0) t1 = INF;
	if (t2 < 0.0) t2 = INF;
	double t = std::min(t1, t2);

	return pos + vel*t;
}

void BotThink() {
	//Moving
	Point radiusvec = POINT(world->player.Position) - Point(0.5, 0.5);
	double dist = !radiusvec;
	if (dist < EPS) radiusvec = Point(1, 0);
	else radiusvec = Normal(radiusvec);
	Point tangentvec(-radiusvec.y, radiusvec.x);
	Point force = tangentvec + radiusvec * (MOVERADIUS - dist) / MOVERADIUS;
	if (dist > 0.4999) dist = 0.4999;
	force += radiusvec * 0.03 / (dist - 0.5);

	for (int i = 0; i<MONSTERS; i++) if (world->monsters[i].ClassName >= 0) {
		Point tovect = POINT(world->monsters[i].Position) - POINT(world->player.Position);
		double dam = world->monsters[i].BiteDamage;
		double dst = !tovect;
		if (dst > EPS && dst < AWARERADIUS) {
			force -= tovect / (dst*dst) * dam * BITEWGT;
		}
	}

	force = Normal(force);
	world->run(COORDS(force));


	//Targeting
	double mind = 1e+50;
	Point best;
	for (int i = 0; i<MONSTERS; i++) if (world->monsters[i].ClassName >= 0) {
		Point tovect = TargetMonsterMoving(world->monsters[i]);
		double score = !tovect;
		if ((force ^ Normal(tovect)) < SHOOTANGLECOS && world->monsters[i].BulletDamage < EPS) score *= 100.0;
		if (mind > score) {
			mind = score;
			best = tovect;
		}
	}
    
	world->shoot(COORDS(best));

}

#ifndef BOTWITHGLUT
World *world;
int main(int argc, char** argv) {
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 666;
	}
	world = new World(argv[1], atoi(argv[2]));
	world->makeMove();
	while (1) {
		BotThink();
		world->makeMove();
	}
	printf("Game over!\n");
	delete world;
	return 0;
}
#endif