#include "../cpp/world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <vector>
#undef min
#undef max
#include <algorithm>
using namespace std;

static const double EPS = 1e-9;
static const double INF = 1e+50;
static const double DELTATIME = 1.0 / 60.0;
static const double PI = 2.0 * acos(0.0);
static const double PIM2 = PI * 2.0;
static const double PID2 = PI / 2.0;

struct Point {
	double x, y;
	
	inline Point() : x(0), y(0) {}
	inline Point(double _x, double _y) : x(_x), y(_y) {}

	inline       double &operator[] (int i)       { return (i==0 ? x : y); }
	inline const double &operator[] (int i) const { return (i==0 ? x : y); }

	inline Point operator+ (const Point &b) const { return Point(x+b.x, y+b.y); }
	inline Point operator- (const Point &b) const { return Point(x-b.x, y-b.y); }
	inline Point operator* (double b) const { return Point(x*b, y*b); }
	inline Point operator/ (double b) const { return *this * (1.0/b); }

	inline void operator+= (const Point &b) { *this = *this + b; }
	inline void operator-= (const Point &b) { *this = *this - b; }
	inline void operator*= (double b) { *this = *this * b; }
	inline void operator/= (double b) { *this = *this / b; }

	inline double operator^ (const Point &b) const { return x*b.x + y*b.y; }
	inline double operator% (const Point &b) const { return x*b.y - y*b.x; }

	inline double operator!() const { return sqrt(*this ^ *this); }
};

Point Normal(const Point &x) {
	Point res = x;
	double norm = !res;
	if (norm > EPS) res /= norm;
	return res;
}

inline double trandom() { return rand() / 32768.0; }

#define POINT(name) Point(name##X, name##Y)
#define COORDS(pnt) (pnt).x, (pnt.y)
static const int MONSTERS = World::MAXMONSTERS;
static const int MISSILES = World::MAXMISSILES;

extern World *world;


static const double MOVERADIUS = 0.4;
static const double SPREADTIME = 0.1;
static const double AWARERADIUS = 0.4;
static const double BITEWGT = 0.5;
static const double SHOOTANGLECOS = 0.70710678118654752440084436210485;
static const double PATHTIME = 5.0;

Point TargetMonsterDull(const CMonster &monster) {
	return POINT(monster.Position) - POINT(world->player.Position);
}

Point TargetMonsterMoving(const CMonster &monster) {
	Point pos = POINT(monster.Position) - POINT(world->player.Position);
	Point vel = POINT(monster.Velocity);
	pos += vel * (2.0*trandom()-1.0) * SPREADTIME;
	double speed = world->player.BulletVelocity;

	double qa = (vel ^ vel) - (speed * speed);
	double qb = (pos ^ vel);
	double qc = (pos ^ pos);
	if (qa > -EPS) return pos;
	double qd = qb*qb - qa*qc;
	qd = sqrt(qd);
	double t1 = (-qb - qd) / qa;
	double t2 = (-qb + qd) / qa;
	if (t1 < 0.0) t1 = INF;
	if (t2 < 0.0) t2 = INF;
	double t = std::min(t1, t2);

	return pos + vel*t;
}

Point ForceToField(Point pos, double rad) {
	if (pos.x < rad) pos.x = rad;
	if (pos.y < rad) pos.y = rad;
	if (pos.x > 1.0-rad) pos.x = 1.0-rad;
	if (pos.y > 1.0-rad) pos.y = 1.0-rad;
	return pos;
}

double MeleeRiskFunc(double dst, double mov, double bor) {
	if (dst <= bor + EPS) return 1.0;
	if (dst >= mov + bor - EPS) return 0.0;
	return 1.0 - (dst - bor) / mov;
}
double GetRisk(Point pos, double dt) {
	double res = 0.0;
	for (int i = 0; i<MONSTERS; i++) {
		const CMonster &mon = world->monsters[i];
		if (mon.ClassName < 0) continue;

		Point mpos = POINT(mon.Position) + POINT(mon.Velocity) * dt;
		double dst = !(mpos - pos);
		double mov = mon.MaxVelocity/2 * dt;
		double bor = mon.Radius + 1.5 * world->player.Radius;

		double mult = EPS;
		if (mon.BiteDamage > EPS) mult = mon.BiteDamage;
		res += MeleeRiskFunc(dst, mov, bor) * mult;
	}
	for (int i = 0; i<MISSILES; i++) {
		const CMissile &miss = world->missiles[i];
		if (miss.ClassName < 0) continue;

		Point mpos = POINT(miss.Position) + POINT(miss.Velocity) * dt;
		double dst = !(mpos - pos);
		double bor = miss.Radius + 1.5 * world->player.Radius;

		double mult = EPS;
		if (miss.Damage > EPS) mult = miss.Damage;
		res += (dst < bor) * mult;
	}

	double bdst = min(min(pos.x, 1.0 - pos.x), min(pos.y, 1.0 - pos.y));
	double warn = 0.9;
	if (bdst > warn) res += 100.0 / ((1.0 - bdst) / (1.0 - warn));

	return res;
}

struct Path {
	int k;
	vector<Point> pos, vel;

	void Resize(int _k) {
		k = _k;
		pos.resize(k+1);
		vel.resize(k+1);
	}

	int Get(double t) {
		t *= k;
		int i = int(t);
		if (i < 0) i = 0;
		if (i > k-1) i = k-1;
		return i;
	}

	Point GetPos(double t) {
		int i = Get(t);
		return pos[i] * (1.0-t) + pos[i+1] * t;
	}
	Point GetVel(double t) {
		int i = Get(t);
		return vel[i];
	}
};

namespace ChooseBigPath {
	static const int PATHBIGPIECES = 6;
	static const int PATHBIGANGLES = 4;

	static const double PATHBIGDT = PATHTIME / PATHBIGPIECES;

	Path best;
	double bestrisk;
	Path bigpath;
	double currentrisk;
	
	void RecPath(int s) {
		double wasrisk = currentrisk;
		currentrisk += GetRisk(bigpath.pos[s], s * PATHBIGDT);
		if (currentrisk < bestrisk) {
		
			if (s == PATHBIGPIECES) {
				if (currentrisk < bestrisk) {
					bestrisk = currentrisk;
					best = bigpath;
				}
			}
			else {
				double offset = trandom() * PIM2;
				for (int i = 0; i<PATHBIGANGLES; i++) {
					double angle = offset + i * (PIM2 / PATHBIGANGLES);
					bigpath.vel[s+1] = Point(cos(angle), sin(angle)) * world->player.MaxVelocity;
					bigpath.pos[s+1] = ForceToField(bigpath.pos[s] + bigpath.vel[s] * PATHBIGDT, world->player.Radius);
					RecPath(s+1);
				}
			}

		}
		currentrisk = wasrisk;
	}
	Path ChoosePath() {
		Point startpos = POINT(world->player.Position);
		Point startvel = POINT(world->player.Velocity);
		bigpath.Resize(PATHBIGPIECES);
		bigpath.pos[0] = startpos;
		bigpath.vel[0] = startvel;
		currentrisk = 0.0;
		bestrisk = INF;
		RecPath(0);
		return best;
	}
}

void BotThink() {
	//Moving
	Path path = ChooseBigPath::ChoosePath();
	Point force = path.vel[1];

	force = Normal(force);
	world->run(COORDS(force));


	//Targeting
	double mind = 1e+50;
	Point best;
	for (int i = 0; i<MONSTERS; i++) if (world->monsters[i].ClassName >= 0) {
		Point tovect = TargetMonsterMoving(world->monsters[i]);
		double score = !tovect;
		if ((force ^ Normal(tovect)) < SHOOTANGLECOS && world->monsters[i].BulletDamage < EPS) score *= 100.0;
		if (mind > score) {
			mind = score;
			best = tovect;
		}
	}
    
	world->shoot(COORDS(best));

}

#ifndef BOTWITHGLUT
World *world;
int main(int argc, char** argv) {
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 666;
	}
	world = new World(argv[1], atoi(argv[2]));
	world->makeMove();
	while (1) {
		BotThink();
		world->makeMove();
	}
	printf("Game over!\n");
	delete world;
	return 0;
}
#endif